require 'json'
require 'ffi'
require 'msgpack'


# all functions which can trigger synchronuous callbacks, and those which wait on rust locks are "blocking: true" to avoid deadlock on GIL
module Hakuban::FFI

	class FFILocalNodeNewResult < FFI::Struct
		layout :error, :uint8, :local_node_pointer, :pointer
	end
	
	class FFIObjectObserveResult < FFI::Struct
		layout :error, :uint8, :object_observe_pointer, :pointer
	end
	
	class FFIObjectExposeResult < FFI::Struct
		layout :error, :uint8, :object_expose_pointer, :pointer
	end
	
	class FFITagObserveResult < FFI::Struct
		layout :error, :uint8, :tag_observe_pointer, :pointer
	end
	
	class FFITagExposeResult < FFI::Struct
		layout :error, :uint8, :tag_expose_pointer, :pointer
	end
	
	class FFIObjectObserveState < FFI::Struct
		layout :synchronized, :uint8, :version_length, :size_t, :version, :pointer, :data_type_length, :size_t, :data_type, :pointer, :raw_length, :size_t, :raw, :pointer, :raw_arc, :pointer
	end
	
	class FFIObjectExposeState < FFI::Struct
		layout :version_length, :size_t, :version, :pointer, :data_type_length, :size_t, :data_type, :pointer, :raw_length, :size_t, :raw, :pointer
		attr_accessor :data_type_strings
	
		def self.construct(version, data_type, data)
			state = FFIObjectExposeState::new
			state[:version_length] = version.size
			state[:version] = FFI::MemoryPointer.new(:int64, version.size)
			state[:version].write_array_of_int64(version)
			state[:data_type_length] = data_type.size
			state[:data_type] = FFI::MemoryPointer.new(:pointer, data_type.size)
			state.data_type_strings = data_type.map {|string| FFI::MemoryPointer.from_string(string)}
			state[:data_type].write_array_of_pointer(state.data_type_strings)
			state[:raw] = FFI::MemoryPointer.from_string(data)
			state[:raw_length] = data.size
			state
		end
	end
	
	class FFIObjectExposeStateResult < FFI::Struct
		layout :error, :uint8, :changed, :uint8
	end
	
	class FFITokioWebsocketConnectorNewResult < FFI::Struct
		layout :error, :uint8, :websocket_connector_pointer, :pointer
	end
	
	class FFITagObserveObjectStateBorrowResult < FFI::Struct
		layout :error, :uint8, :state, FFIObjectObserveState.by_value
	end
	
	class FFIObjectDescriptor < FFI::Struct
		layout :tags_count, :size_t, :tags, :pointer, :json, :pointer
		attr_accessor :tags_strings
		
		def tags
			self[:tags].read_array_of_pointer(self[:tags_count]).map { |string| JSON.parse(string.read_string()) }	# does this copy the string? 
		end
	
		def json
			JSON.parse(self[:json].read_string())
		end
	
		def self.construct(tags,json)
			descriptor = FFIObjectDescriptor.new
			descriptor[:tags_count] = tags.size
			descriptor.tags_strings = tags.map { |tag| FFI::MemoryPointer.from_string(JSON.dump(tag)) }
			descriptor[:tags] = FFI::MemoryPointer.new(:pointer, descriptor.tags_strings.size)
			descriptor[:tags].write_array_of_pointer(descriptor.tags_strings)
			descriptor[:json] = FFI::MemoryPointer.from_string(JSON.dump(json))
			descriptor
		end
	
	end
	
	class FFITagDescriptor < FFI::Struct
		layout :json, :pointer
	
		def json
			JSON.parse(self[:json].read_string())
		end
	
		def self.construct(json)
			descriptor = FFITagDescriptor.new
			descriptor[:json] = FFI::MemoryPointer.from_string(JSON.dump(json))
			descriptor
		end
	
	end
	
	
	class FFIObjectDescriptors < FFI::Struct
		layout :count, :size_t, :descriptors, :pointer
	
		def descriptors
			self[:count].times.map { |i| FFIObjectDescriptor.new(self[:descriptors] + i * FFIObjectDescriptor.size) }
		end
	end
	

	extend FFI::Library

	ffi_lib 'hakuban'

	callback :object_callback, [ :pointer, FFIObjectDescriptor.by_value, :uint8 ], :void
	#callback :tag_callback, [ :pointer, FFITagDescriptor.by_value, :uint8 ], :void

	attach_function :hakuban_local_node_new, [ :string ], FFILocalNodeNewResult.by_value
	attach_function :hakuban_local_node_drop, [ :pointer ], :void, blocking: true

	attach_function :hakuban_object_observe_new, [ :pointer, FFIObjectDescriptor.by_value ], FFIObjectObserveResult.by_value, blocking: true
	attach_function :hakuban_object_observe_drop, [ :pointer ], :void, blocking: true
	attach_function :hakuban_object_observe_state_borrow, [ :pointer ], FFIObjectObserveState.by_value, blocking: true
	attach_function :hakuban_object_observe_changes_callback_register, [ :pointer, :object_callback, :pointer ], :pointer, blocking: true

	attach_function :hakuban_object_expose_new, [ :pointer, FFIObjectDescriptor.by_value], FFIObjectExposeResult.by_value, blocking: true
	attach_function :hakuban_object_expose_drop, [ :pointer ], :void, blocking: true
	attach_function :hakuban_object_expose_state, [ :pointer, FFIObjectExposeState.by_value ], FFIObjectExposeStateResult.by_value, blocking: true
	attach_function :hakuban_object_expose_assigned, [ :pointer ], :uint8, blocking: true
	attach_function :hakuban_object_expose_changes_callback_register, [ :pointer, :object_callback, :pointer ], :pointer, blocking: true

	attach_function :hakuban_tag_observe_new, [ :pointer, FFITagDescriptor.by_value ], FFITagObserveResult.by_value, blocking: true
	attach_function :hakuban_tag_observe_drop, [ :pointer ], :void, blocking: true
	attach_function :hakuban_tag_observe_object_descriptors_borrow, [ :pointer ], FFIObjectDescriptors.by_value, blocking: true
	attach_function :hakuban_tag_observe_object_state_borrow, [ :pointer, FFIObjectDescriptor.by_value ], FFITagObserveObjectStateBorrowResult.by_value, blocking: true
	attach_function :hakuban_tag_observe_objects_changes_callback_register, [ :pointer, :object_callback, :pointer ], :pointer, blocking: true

	attach_function :hakuban_tag_expose_new, [ :pointer, FFITagDescriptor.by_value ], FFITagExposeResult.by_value, blocking: true
	attach_function :hakuban_tag_expose_drop, [ :pointer ], :void, blocking: true
	attach_function :hakuban_tag_expose_object_descriptors_borrow, [ :pointer ], FFIObjectDescriptors.by_value, blocking: true
	attach_function :hakuban_tag_expose_object_state, [ :pointer, FFIObjectDescriptor.by_value, FFIObjectExposeState.by_value ], FFIObjectExposeStateResult.by_value, blocking: true
	attach_function :hakuban_tag_expose_objects_changes_callback_register, [ :pointer, :object_callback, :pointer ], :pointer, blocking: true

	attach_function :hakuban_object_observe_state_return, [ FFIObjectObserveState.by_value ], :void
	attach_function :hakuban_object_descriptors_return, [ FFIObjectDescriptors.by_value ], :void
	attach_function :hakuban_object_callback_unregister, [ :pointer ], :void, blocking: true

	attach_function :hakuban_tokio_init_multi_thread, [ :size_t ], :pointer
	attach_function :hakuban_tokio_websocket_connector_new, [ :pointer, :pointer, :string ], FFITokioWebsocketConnectorNewResult.by_value
	attach_function :hakuban_tokio_websocket_connector_drop, [ :pointer, :pointer ], :void, blocking: true

	attach_function :hakuban_logger_initialize, [ :string ], :uint8

end
