require 'set'
require 'ostruct'

require_relative './ffi.rb'

#TODO: explicit drops?
#TODO: privative methods
#TODO: error classes


module Hakuban

	def self.raise_if_error(result)
		case result[:error] 
		when 0 then true
		when 1 then raise "Invalid string"
		when 2 then raise "Invalid JSON"
		when 3 then raise "Invalid URL"
		when 4 then raise "Object Not Found"
		else raise "Unknown error, sorry :("
		end
	end


	def self.action_int_to_symbol(number)
		case number
		when 0 then :insert
		when 1 then :change
		when 2 then :remove
		end
	end


	@@logger_initialized = false

	def self.logger_initialize(default_level, skip_if_already_initialized = false)
		if @@logger_initialized and !skip_if_already_initialized
			raise "Logger already initialized. This can't be done more than once. Make sure logger_initialize is called before any LocalNode gets constructed."  
		end
		if not @@logger_initialized
			raise "Invalid default log level string"  if FFI::hakuban_logger_initialize(default_level) != 0
			@@logger_initialized = true
		end
	end


	class LocalNode

		#TODO: explicit drop

		attr_reader :local_node_pointer  #todo: hide

		def initialize(name=nil)
			@default_serializer = lambda { |data_type, data| 
				[["JSON"]+data_type, JSON.dump(data)] 
			}
			@default_deserializer = lambda { |data_type, data| 
				raise "Expected JSON serialized data, got: #{dat_type}"  if data_type[0] != "JSON"
				[data_type[1..-1], JSON.load(data)] 
			}
			Hakuban::logger_initialize("hakuban=warn", true)
			result = FFI::hakuban_local_node_new(name || File.basename(caller_locations(0..1)[1].path))
			Hakuban::raise_if_error(result)
			@local_node_pointer = ::FFI::AutoPointer.new(result[:local_node_pointer], FFI::method(:hakuban_local_node_drop))
		end

		def with_default_serializer(&block)
			@default_serializer = block
			self
		end

		def with_default_deserializer(&block)
			@default_deserializer = block
			self
		end

		def default_serializer(&block)
			if block_given?
				@default_serializer = block
			else
				@default_serializer
			end
		end

		def default_deserializer(&block)
			if block_given?
				@default_deserializer = block
			else
				@default_deserializer
			end
		end

		def object(tags, descriptor)
			#TODO: accept real descriptor too
			ObjectBuilder.new(self, ObjectDescriptor.new(tags,descriptor), @default_serializer, @default_deserializer)
		end

		def tag(descriptor)
			#TODO: accept real descriptor too
			TagBuilder.new(self, TagDescriptor.new(descriptor), @default_serializer, @default_deserializer)
		end

	end


	class ObjectBuilder

		def initialize(store, descriptor, serializer, deserializer)
			@store, @descriptor, @serializer, @deserializer = store, descriptor, serializer, deserializer
		end

		def observe
			ObjectObserve.new(@store, @descriptor, @deserializer)
		end

		def expose
			ObjectExpose.new(@store, @descriptor, @serializer)
		end

		def with_serializer(&block)
			@serializer = block
		end

		def with_deserializer(&block)
			@deserializer = block
		end

	end


	class TagBuilder

		def initialize(store, descriptor, serializer, deserializer)
			@store, @descriptor, @serializer, @deserializer = store, descriptor, serializer, deserializer
		end

		def observe
			TagObserve.new(@store, @descriptor, @deserializer)
		end

		def expose
			TagExpose.new(@store, @descriptor, @serializer)
		end

		def with_serializer(&block)
			@serializer = block
		end

		def with_deserializer(&block)
			@deserializer = block
		end

	end


	class ObjectDescriptor
		attr_reader :tags, :json

		def initialize(tags,json)
			@tags, @json = Set.new(tags), json
		end

		def to_ffi
			FFI::FFIObjectDescriptor.construct(@tags, @json)
		end

		def self.from_ffi(ffi)
			ObjectDescriptor.new(ffi.tags, ffi.json)
		end

		def ==(other)
			@tags == other.tags and @json == other.json
		end

		alias eql? ==
		
		def hash
			[@tags.hash, @json.hash].hash
		end

		def inspect
			"#<ObjectDescriptor @tags={%s}, @json=%p>"%[self.tags.map(&:inspect).join(","), self.json]
		end

	end
	

	class TagDescriptor
		attr_reader :json

		def initialize(json)
			@json = json
		end

		def to_ffi
			FFI::FFITagDescriptor.construct(@json)
		end

		def self.from_ffi(ffi)
			TagDescriptor.new(ffi.json)
		end

		def ==(other)
			@json == other.json
		end

		alias eql? ==

		def hash
			@json.hash
		end

	end


	class EventQueue

		def initialize(queue, prefix, register_method, pointer, unregister_method)
			@queue = queue
			@callback = EventQueue.generate_callback(@queue, prefix)
			@registered_callback_pointer = ::FFI::AutoPointer.new(
				register_method.call(pointer, @callback, ::FFI::Pointer::NULL), 
				EventQueue.generate_unregister(unregister_method, @callback)
			)
			#puts "Allocated callback pointer: #{@registered_callback_pointer}"
			#puts "Allocated callback queue: #{@events}"
		end	

		def self.generate_callback(queue, prefix)
			proc { |_userdata, ffi_descriptor, ffi_action|
				#puts "Callback for queue: #{queue}"
				action = Hakuban::action_int_to_symbol(ffi_action)
				descriptor = ObjectDescriptor.from_ffi(ffi_descriptor)
				queue << prefix + [action, descriptor]
			}
		end

		def self.generate_unregister(unregister_method, callback)
			proc { |pointer|
				#puts "Dropping callback pointer: #{pointer}"
				callback # probably not needed, but I want to be sure this is captured
				unregister_method.call(pointer)
			}
		end

	end


	class ObjectObserve

		attr_reader :descriptor

		def initialize(local_node, descriptor, deserializer)
			@local_node, @descriptor, @deserializer = local_node, descriptor, deserializer
			result = FFI::hakuban_object_observe_new(@local_node.local_node_pointer, descriptor.to_ffi)
			Hakuban::raise_if_error(result)
			@object_observe_pointer = ::FFI::AutoPointer.new(result[:object_observe_pointer], FFI::method(:hakuban_object_observe_drop))
			@queues = {}
		end

		def object_state
			raise "Attempt to use after 'drop'"  if not @object_observe_pointer
			ObjectObserveState.new(FFI::hakuban_object_observe_state_borrow(@object_observe_pointer), @deserializer)
		end

		def send_events_to(queue, *prefix)
			@queues[queue] = EventQueue.new(
				queue,
				prefix,
				FFI::method(:hakuban_object_observe_changes_callback_register), 
				@object_observe_pointer, 
				FFI::method(:hakuban_object_callback_unregister)
			)
		end

		def stop_sending_objects_events_to(queue)
			@queues.delete(queue)
		end

		def inspect
			"#<ObjectObserve #{@descriptor}>"
		end

		def drop
			@object_observe_pointer.free
			@object_observe_pointer = nil
			@queues.clear
		end

		def manage(manager=nil)
			manager ||= ObjectManager.new
			manager.contract = self
			manager
		end

	end


	class ObjectExpose

		attr_reader :descriptor

		def initialize(local_node, descriptor, serializer)
			@local_node, @descriptor, @serializer = local_node, descriptor, serializer
			result = FFI::hakuban_object_expose_new(@local_node.local_node_pointer, descriptor.to_ffi)
			Hakuban::raise_if_error(result)
			@object_expose_pointer = ::FFI::AutoPointer.new(result[:object_expose_pointer], FFI::method(:hakuban_object_expose_drop))
			@queues = {}
		end

		def set_object_state(version, data, data_type = [])
			raise "Attempt to use after 'drop'"  if not @object_expose_pointer
			serialized_data_type, serialized_data = @serializer.call(data_type, data)
			result = FFI::hakuban_object_expose_state(@object_expose_pointer, FFI::FFIObjectExposeState.construct(version, serialized_data_type, serialized_data))
			Hakuban::raise_if_error(result)
			result[:changed] == 1
		end

		def assigned
			raise "Attempt to use after 'drop'"  if not @object_expose_pointer
			FFI::hakuban_object_expose_assigned(@object_expose_pointer) == 1
		end

		def send_events_to(queue, *prefix)
			@queues[queue] = EventQueue.new(
				queue,
				prefix,
				FFI::method(:hakuban_object_expose_changes_callback_register), 
				@object_expose_pointer,
				FFI::method(:hakuban_object_callback_unregister)
			)
		end

		def stop_sending_objects_events_to(queue)
			@queues.delete(queue)
		end

		def inspect
			"#<ObjectExpose #{@descriptor}>"
		end

		def drop
			@object_expose_pointer.free
			@object_expose_pointer = nil
			@queues.clear
		end

		def manage(manager=nil)
			manager ||= ObjectManager.new
			manager.contract = self
			manager
		end

	end


	class ObjectObserveState

		attr_reader :data, :data_type, :data_version, :synchronized

		def initialize(raw_state, deserializer)
			@raw_state = raw_state
			@synchronized = (raw_state[:synchronized] == 1)
			@data_version = raw_state[:version].read_array_of_type(::FFI::TYPE_INT64, :read_int64, raw_state[:version_length])  if not raw_state[:version].null?
			serialized_data_type = raw_state[:data_type].read_array_of_pointer(raw_state[:data_type_length]).map { |string| string.read_string() }
			@data_type, @data = deserializer.call(serialized_data_type, raw_state[:raw].read_string_length(raw_state[:raw_length]))  if not raw_state[:raw].null?
			ObjectSpace.define_finalizer(self, ObjectObserveState.finalize(@raw_state))
		end

		def ObjectObserveState.finalize(raw_state)
			proc { |_|
				FFI::hakuban_object_observe_state_return(raw_state)
			}
		end

		def version
			@data_version
		end

		def inspect
			"#<ObjectObserveState @synchronized=%p @version=%p>"%[@synchronized, @data_version]
		end

	end


	class TagObserve

		attr_reader :descriptor

		def initialize(local_node, descriptor, deserializer)
			@local_node, @descriptor, @deserializer = local_node, descriptor, deserializer
			result = FFI::hakuban_tag_observe_new(@local_node.local_node_pointer, @descriptor.to_ffi)
			Hakuban::raise_if_error(result)
			@tag_observe_pointer = ::FFI::AutoPointer.new(result[:tag_observe_pointer], FFI::method(:hakuban_tag_observe_drop))
			@queues = {}
		end

		def object_descriptors
			raise "Attempt to use after 'drop'"  if not @tag_observe_pointer
			result = FFI::hakuban_tag_observe_object_descriptors_borrow(@tag_observe_pointer)
			ret = result.descriptors.map { |raw_descriptor| ObjectDescriptor::from_ffi(raw_descriptor) }
			FFI::hakuban_object_descriptors_return(result)
			ret
		end
		
		def object_state(object_descriptor)
			raise "Attempt to use after 'drop'"  if not @tag_observe_pointer
			result = FFI::hakuban_tag_observe_object_state_borrow(@tag_observe_pointer, object_descriptor.to_ffi)
			return nil  if result[:error] == 4
			Hakuban::raise_if_error(result)
			ObjectObserveState.new(result[:state], @deserializer)
		end

		def send_events_to(*params)
			self.send_objects_events_to(*params)
		end

		def send_objects_events_to(queue, *prefix)
			@queues[queue] = EventQueue.new(
				queue,
				prefix,
				FFI::method(:hakuban_tag_observe_objects_changes_callback_register), 
				@tag_observe_pointer, 
				FFI::method(:hakuban_object_callback_unregister)
			)
		end

		def stop_sending_objects_events_to(queue)
			@queues.delete(queue)
		end

		def drop
			@tag_observe_pointer.free
			@tag_observe_pointer = nil
			@queues.clear
		end

		def manage(manager=nil)
			manager ||= ObjectManager.new
			manager.contract = self
			manager
		end

	end


	class TagExpose

		attr_reader :descriptor
		
		def initialize(local_node, descriptor, serializer)
			@local_node, @descriptor, @serializer = local_node, descriptor, serializer
			result = FFI::hakuban_tag_expose_new(@local_node.local_node_pointer, descriptor.to_ffi)
			Hakuban::raise_if_error(result)
			@tag_expose_pointer = ::FFI::AutoPointer.new(result[:tag_expose_pointer], FFI::method(:hakuban_tag_expose_drop))
			@queues = {}
		end

		def object_descriptors
			raise "Attempt to use after 'drop'"  if not @tag_expose_pointer
			result = FFI::hakuban_tag_expose_object_descriptors_borrow(@tag_expose_pointer)
			ret = result.descriptors.map { |raw_descriptor| ObjectDescriptor::from_ffi(raw_descriptor) }
			FFI::hakuban_object_descriptors_return(result)
			ret
		end
		
		def set_object_state(object_descriptor, version, data, data_type = [])
			raise "Attempt to use after 'drop'"  if not @tag_expose_pointer
			serialized_data_type, serialized_data = @serializer.call(data_type, data)
			result = FFI::hakuban_tag_expose_object_state(@tag_expose_pointer, object_descriptor.to_ffi, FFI::FFIObjectExposeState.construct(version, serialized_data_type, serialized_data))
			Hakuban::raise_if_error(result)
			result[:changed] == 1
		end

		def send_events_to(*params)
			self.send_objects_events_to(*params)
		end

		def send_objects_events_to(queue, *prefix)
			@queues[queue] = EventQueue.new(
				queue,
				prefix,
				FFI::method(:hakuban_tag_expose_objects_changes_callback_register), 
				@tag_expose_pointer, 
				FFI::method(:hakuban_object_callback_unregister)
			)
		end

		def stop_sending_objects_events_to(queue)
			@queues.delete(queue)
		end

		def drop
			@tag_expose_pointer.free
			@tag_expose_pointer = nil
			@queues.clear
		end

		def manage(manager=nil)
			manager ||= ObjectManager.new
			manager.contract = self
			manager
		end

	end



	class Tokio

		@@tokio_pointer = nil

		def Tokio.init(workers_count=0)
			@@tokio_pointer ||= FFI::hakuban_tokio_init_multi_thread(0)
		end

		def Tokio.pointer
			Tokio.init  if not @@tokio_pointer
			@@tokio_pointer
		end

		#TODO: drop?	

	end


	class WebsocketConnector

		#TODO: drop	

		def initialize(local_node, url)
			result = FFI::hakuban_tokio_websocket_connector_new(Tokio.pointer, local_node.local_node_pointer, url)
			Hakuban::raise_if_error(result)
			@websocket_connector_pointer = ::FFI::AutoPointer.new(result[:websocket_connector_pointer], WebsocketConnector.generate_drop)
		end

		def self.generate_drop
			proc { |pointer|
				FFI::hakuban_tokio_websocket_connector_drop(Tokio.pointer, pointer)
			}
		end

	end


	# convenience utils

	#TODO: replace this mess (<Hash, dynamic mixins) with some "deref"
	class ObjectManager < Hash

		module ObservedObjectManager
			def descriptor;                             self.values.first.descriptor;              end
			def state;                                  self.values.first.state;                   end
			def data;                                   self.values.first.data;                    end
			def version;                                self.values.first.verison;                 end
			def synchronized;                           self.values.first.synchronized;            end
			def construct_object(contract, descriptor); ObservedObject.new(contract, descriptor);  end
		end

		module ExposedObjectManager
			def descriptor;                             self.values.first.descriptor;              end
			def state=(params);                         self.values.first.state = params;          end
			def data=(params);                          self.values.first.data = params;           end
			def state_set(params);                      self.values.first.state_set(params);       end
			def data_set(params);                       self.values.first.data_set(params);        end
			def construct_object(contract, descriptor); ExposedObject.new(contract, descriptor);   end
		end

		module ObservedTagManager
			def construct_object(contract, descriptor); ObservedObject.new(contract, descriptor);  end
		end

		module ExposedTagManager
			def construct_object(contract, descriptor); ExposedObject.new(contract, descriptor);  end
		end
		
		attr_reader :objects, :contract

		def contract=(contract)
			raise "You can't change contracts, sorry."  if @contract
			@contract = contract
			case @contract.class.name
				when "Hakuban::ObjectObserve" then singleton_class.include(ObservedObjectManager)
				when "Hakuban::ObjectExpose"  then singleton_class.include(ExposedObjectManager)
				when "Hakuban::TagObserve" then singleton_class.include(ObservedTagManager)
				when "Hakuban::TagExpose" then singleton_class.include(ExposedTagManager)
				else raise
			end

			@queue = Queue.new
			@contract.send_events_to(@queue, :object)
			if @contract.kind_of?(ObjectObserve) or @contract.kind_of?(ObjectExpose)
				event, param, descriptor = @queue.pop				 
				raise if event != :object or param != :insert or not descriptor
				self[descriptor] = self.construct_object(@contract, descriptor)
			end

			@thread = Thread.new {
				loop {
					event,param,descriptor = @queue.shift
					case [event, param]
					when [:control, :quit]
						break
					when [:object, :insert]
						self[descriptor] = construct_object(@contract, descriptor)
						object_insert(self[descriptor])
					when [:object, :change]
						object_change(self[descriptor])
					when [:object, :remove]
						object_remove(delete(descriptor))
					end
				}
			}
		end


		def object_insert(object)
			# noop default
		end

		def object_change(object)
			# noop default
		end

		def object_remove(object)
			# noop default
		end


		def drop
			@contract.stop_sending_objects_events_to(@queue)
			@queue.push [:control, :quit]
			@thread.join
			@queue, @contract, @thread, @objects = nil, nil, nil, {}
		end


		class ManagedObject < OpenStruct
			attr_reader :contract, :descriptor
			def initialize(contract, descriptor)
				@contract, @descriptor = contract, descriptor
				super()
			end
		end	

		#TODO: remove the arity thing...
		class ObservedObject < ManagedObject
			def state
				if @contract.method(:object_state).arity == 1
					@contract.object_state(@descriptor)
				else
					@contract.object_state
				end
			end
			def data
				if @contract.method(:object_state).arity == 1
					@contract.object_state(@descriptor)&.data
				else
					@contract.object_state&.data
				end
			end
		end

		class ExposedObject < ManagedObject
			def state_set(state)
				if @contract.method(:set_object_state).arity == -4
					@contract.set_object_state(@descriptor, *state)
				else
					@contract.set_object_state(*state)
				end
			end
			def state=(state)
				state_set(state)
			end
			def data_set(value)
				timestamp = Time.new.to_f
				if @contract.method(:set_object_state).arity == -4
					@contract.set_object_state(@descriptor,[1, timestamp.floor, ((timestamp - timestamp.floor)*1000000000).floor, 0],value)
				else 
					@contract.set_object_state([1, timestamp.floor, ((timestamp - timestamp.floor)*1000000000).floor, 0],value)
				end
			end
			def data=(value)
				data_set(value)
			end
		end
	end




end